#!/bin/env python

__author__ = "John Hover"
__copyright__ = "2017 John Hover"
__credits__ = []
__license__ = "GPL"
__version__ = "0.9.1"
__maintainer__ = "John Hover"
__email__ = "jhover@bnl.gov"
__status__ = "Production"

import logging

class InfoEntity(object):
    '''
    Template for Information entities. Common functions. 
    Classes that inherit from InfoEntity must set class variables to describe handling. 

    '''
    infokey = 'unset'
    infoattributes = []
    validvalues = {}
        
    def __repr__(self):
        s = "%s(" % self.__class__.__name__
        for a in self.__class__.infoattributes:
            s+="%s=%s " % (a, getattr(self, a, None)) 
        s += ")"
        return s    

    def makeDictObject(self):
        '''
        Converts this Python object to attribute dictionary suitable for addition to existing dict 
        intended to be converted back to JSON. Uses <obj>.name as key:
        '''
        d = {}
        d[self.name] = {}
        for attrname in self.infoattributes:
            d[self.name][attrname] = getattr(self, attrname)
        self.log.debug("Returning dict: %s" % d)
        return d    

    def setState(self, newstate):
        self.log.debug("%s object name=%s %s ->%s" % (self.__class__.__name__, self.name, self.state, newstate) )
        self.state = newstate
    
    def store(self, infoclient):
        '''
        Stores this Infoentity in the provided infoclient info tree. 
        '''
        keystr = self.__class__.infokey
        validvalues = self.__class__.validvalues
        for keyattr in validvalues.keys():
            validlist = validvalues[keyattr]
            attrval = getattr(self, keyattr) 
            if attrval not in validlist:
                self.log.warning("Info Entity has invalid value '%s' for attribute '%s' " % (attrval, keyattr) )
        #resources = infoclient.getdocumentobject(key=keystr)
        da = self.makeDictObject()
        self.log.debug("Dict obj: %s" % da)
        infoclient.storedocumentobject(da, key=keystr)

    def addAcl(self, aclstring):
        pass    

    def removeAcl(self, aclstring):
        pass

    @classmethod
    def objectFromDict(cls, dict):
        '''
        Returns an initialized Entity object from dictionary. 
        Input: Dict:
        { <name> : 
            {
                "name" : "<name>",
                "att1" : "<val1>"  
            }
        }
        '''
        log = logging.getLogger()
        log.debug("Making object from dictionary...")
        name = dict.keys()[0]
        d = dict[name]
        args = {}
        for key in cls.infoattributes:
            args[key] = d[key]
        eo = cls(**args)
        log.debug("Successfully made object from dictionary, returning...")
        return eo


#
# __init__.py
#
#
"""
Generic __init__.py to allow 

        from Package.path import *

to work correctly. 

"""
from os.path import basename, dirname, splitext
import glob

allmods = []

leftpath = dirname(__file__)

for filename in glob.glob("%s/*.py" % leftpath):
    if filename.find("__") < 0:   # ignore __init__.py files
        #print "python module file: %s " % filename
        modfilename = basename(filename)
        modname = modname, ext = splitext(modfilename)
        allmods.append(modname)

# List of all module names valid to import from this package
__all__= allmods


#! /usr/bin/env python
__author__ = "John Hover, Jose Caballero"
__copyright__ = "2017 John Hover"
__credits__ = []
__license__ = "GPL"
__version__ = "0.9.1"
__maintainer__ = "John Hover"
__email__ = "jhover@bnl.gov"
__status__ = "Production"

# utilites for devel, debug
# consume python config files, interpolate, emit human-READABLE
# consume file and urlencode

class UtilCLI(object):
    def __init__(self):
        pass


if __name__ == "__main__":
    pass





#! /usr/bin/env python
__author__ = "John Hover, Jose Caballero"
__copyright__ = "2017 John Hover"
__credits__ = []
__license__ = "GPL"
__version__ = "0.9.1"
__maintainer__ = "John Hover"
__email__ = "jhover@bnl.gov"
__status__ = "Production"


import cherrypy
import logging
import logging.handlers
import os
import platform
import pwd
import random
import json
import string
import socket
import sys
import threading
import time
import traceback

from optparse import OptionParser
from ConfigParser import ConfigParser


# Since script is in package "vc3" we can know what to add to path for 
# running directly during development
(libpath,tail) = os.path.split(sys.path[0])
sys.path.append(libpath)

import pluginmanager as pm

class InfoHandler(object):
    '''
    Handles low-level operations and persistence of information 
    from service using back-end plugin.
    
    Data at this level is converted to/from native Python objects for storage/retrieval
     
    '''
    def __init__(self, config):
        self.log = logging.getLogger()
        self.log.debug("Initializing Info Handler...")
        self.config = config
        
        # Get persistence plugin
        pluginname = config.get('persistence','plugin')
        psect = "plugin-%s" % pluginname.lower()
        self.log.debug("Creating persistence plugin...")
        self.persist = pm.getplugin(parent=self, 
                                    paths=['vc3infoservice', 'plugins', 'persist'], 
                                    name=pluginname, 
                                    config=self.config, 
                                    section=psect)
        self.log.debug("Done initting InfoHandler")

    def storedocument(self, key, doc):
        self.log.debug("Storing document for key %s" % key)
        pd = json.loads(doc)
        self.persist.storedocument(key, pd)
   
    def mergedocument(self, key, doc):
        self.log.debug("input doc to merge is type %s" % type(doc))
        dcurrent = self.persist.getdocument(key)
        self.log.debug("current retrieved doc is type %s" % type(dcurrent))
        md = json.loads(doc)
        self.log.debug("doc to merge is type %s" % type(md))
        newdoc = self.merge( md, dcurrent)
        self.log.debug("Merging document for key %s" % key)
        self.persist.storedocument(key, newdoc)
    
    def getdocument(self, key):
        '''
        Gets JSON representation of document. 
        '''
        pd = self._getpythondocument(key)
        jd = json.dumps(pd)
        self.log.debug("d is type %s" % type(jd))
        return jd

    def _getpythondocument(self, key):
        '''
        Gets Python object. 
        '''
        d = self.persist.getdocument(key)
        self.log.debug("d is type %s with value %s" % (type(d), d))
        return d
    
    def _storepythondocument(self, key, pd):
        self.log.debug("Storing document for key %s" % key)
        self.persist.storedocument(key, pd)
    
    
    def deletesubtree(self, path):
        lst = path.split('.')
        try:
            self.persist.deletesubtree(lst)
        except IndexError:
            raise Exception('path should have more than one key')

    def merge(self, source, destination):
        """
        merges nested python dictionaries.
        run me with nosetests --with-doctest file.py
    
        >>> a = { 'first' : { 'all_rows' : { 'pass' : 'dog', 'number' : '1' } } }
        >>> b = { 'first' : { 'all_rows' : { 'fail' : 'cat', 'number' : '5' } } }
        >>> merge(b, a) == { 'first' : { 'all_rows' : { 'pass' : 'dog', 'fail' : 'cat', 'number' : '5' } } }
        True
        """
        self.log.debug("source is type %s" % type(source))
        self.log.debug("destination is type %s" % type(destination))
        for key, value in source.items():
            self.log.debug("processing node %s" % key)
            if isinstance(value, dict):
                # get node or create one
                node = destination.setdefault(key, {})
                self.merge(value, node)
            else:
                destination[key] = value
        return destination

    ##################################################################################
    #   Pairing-related calls. 
    ##################################################################################
    
    def getpairing(self, key, pairingcode):
        '''
        Pull pairing document, check each entry to see if <entry>.pairingcode = pairingcode.
        If so, and cert and key are not none, prepare to return them, delete entry, return Pairing
        '''
        failmsg="Invalid pairing code or not satisfied yet. Try in 30 seconds."
        prd = None
        pd = self._getpythondocument(key)
        self.log.debug("Received dict: %s" % pd)
        try:        
            self.log.debug("Entries are %s" % pd[key] )
            for p in pd[key].keys():
                self.log.debug("Checking entry %s for pairingcode..." % p)
                if pd[key][p]['pairingcode'] == pairingcode:
                    self.log.debug("Found matching entry %s value %s" % (p, pd[key][p]))
                    if pd[key][p]['cert'] is not None:
                        prd = json.dumps(pd[key][p])
                        try:
                            self.log.debug("Attempting to delete entry %s from pairing." % p)
                            pd[key].pop(p, None)
                            self.log.debug("Deleted entry %s from pairing. Re-storing.." % p)
                        except KeyError:
                            self.log.warning("Failed to delete entry %s from pairing." % p)
                        self._storepythondocument(key, pd)
                    else:
                        self.log.info("Certificate for requested pairing not generated yet.")
            self.log.debug("Returning pairing entry JSON %s" % prd)
            if prd is None:
                cherrypy.response.headers["Status"] = "404"
                return failmsg
            return prd
        except KeyError:
            cherrypy.response.headers["Status"] = "404"
            return failmsg
   
   

class InfoRoot(object):

    @cherrypy.expose
    def index(self):
        return "Nothing to see. Go to /info"    

    @cherrypy.expose
    def generate(self, length=8):
        return ''.join(random.sample(string.hexdigits, int(length)))

class InfoServiceAPI(object):
    ''' 
        Data at this level is assumed to be  JSON text/plain.
    '''
    exposed = True 
    
    def __init__(self, config):
        self.log = logging.getLogger()
        self.log.debug("Initting InfoServiceAPI...")
        self.infohandler = InfoHandler(config)
        self.log.debug("InfoServiceAPI init done." )
    
    def GET(self, key, pairingcode=None):

        self.log.info("CLIENT HEADERS")
        self.log.info(str(cherrypy.request.headers))


        if pairingcode is None:
            d = self.infohandler.getdocument(key) 
            self.log.debug("Document retrieved for key %s with val %s" % (key,d))
            return d
        else:
            self.log.debug("Handling pairing retrieval")
            d = self.infohandler.getpairing(key, pairingcode)
            self.log.debug("Pairing retrieved for code %s with val %s" % (pairingcode,d))
            return d

    @cherrypy.tools.accept(media='text/plain')
    def PUT(self, key, data):
        self.log.debug("Storing document %s" % data)
        self.infohandler.mergedocument(key, data)
        self.log.debug("Document stored for key %s" % key)
        return "Document stored for key %s\n" % key
        
    def POST(self, key, data):
        self.log.debug("Storing document %s" % data)
        self.infohandler.storedocument(key, data)
        self.log.debug("Document stored for key %s" % key)
        return "Document stored for key %s\n" % key
        
    def DELETE(self, key, name):
        '''
        Unusual call that operates on node provided in path only. Path expressed as node 'name' attribute. 
        /info/<key>/<name>/
        Deletion only occurs if identity is allowed delete by ACL at <name>: { 'acl' :<acl> } 
        '''

        self.log.debug("Deleting subtree %s" % name)
        self.infohandler.deletesubtree(name)
        self.log.debug("Subtree deleted at %s" % name)
        return "Subtree deleted at %s" % name


    def stripquotes(self,s):
        rs = s.replace("'","")
        return rs


class InfoService(object):
    
    def __init__(self, config):
        self.log = logging.getLogger()
        self.log.debug('InfoService class init...')
        self.config = config
        self.certfile = os.path.expanduser(config.get('netcomm','certfile'))
        self.keyfile = os.path.expanduser(config.get('netcomm', 'keyfile'))
        self.chainfile = os.path.expanduser(config.get('netcomm','chainfile'))
        self.httpport = int(config.get('netcomm','httpport'))
        self.httpsport = int(config.get('netcomm','httpsport'))
        self.sslmodule = config.get('netcomm','sslmodule')
        
        self.log.debug("certfile=%s" % self.certfile)
        self.log.debug("keyfile=%s" % self.keyfile)
        self.log.debug("chainfile=%s" % self.chainfile)
        
        self.log.debug('InfoService class done.')
        
    def run(self):
        self.log.debug('Infoservice running...')
          
        cherrypy.tree.mount(InfoRoot())
        cherrypy.tree.mount(InfoServiceAPI(self.config),'/info',
                                {'/':
        {'request.dispatch': cherrypy.dispatch.MethodDispatcher()}
    })
        #cherrypy.tree.mount(InfoServiceAPI(self.config))
        
        
        cherrypy.server.unsubscribe()
    
        server1 = cherrypy._cpserver.Server()
        server1.socket_port=self.httpsport
        server1._socket_host='0.0.0.0'
        server1.thread_pool=30
        server1.ssl_module = self.sslmodule
        server1.ssl_certificate = self.certfile
        server1.ssl_private_key = self.keyfile
        server1.ssl_certificate_chain = self.chainfile
        server1.subscribe()
    
        #server2 = cherrypy._cpserver.Server()
        #server2.socket_port=self.httpport
        #server2._socket_host="0.0.0.0"
        #server2.thread_pool=30
        #server2.subscribe()
    
        cherrypy.engine.start()
        cherrypy.engine.block()   
    

class InfoServiceCLI(object):
    """class to handle the command line invocation of service. 
       parse the input options,
       setup everything, and run InfoService class
    """
    def __init__(self):
        self.options = None 
        self.args = None
        self.log = None
        self.config = None

        self.__presetups()
        self.__parseopts()
        self.__setuplogging()
        self.__platforminfo()
        self.__checkroot()
        self.__createconfig()

    def __presetups(self):
        '''
        we put here some preliminary steps that 
        for one reason or another 
        must be done before anything else
        '''

    
    def __parseopts(self):
        parser = OptionParser(usage='''%prog [OPTIONS]
vc3-info-service is a information store for VC3

This program is licenced under the GPL, as set out in LICENSE file.

Author(s):
John Hover <jhover@bnl.gov>
''', version="%prog $Id: infoservice.py 1-13-17 23:58:06Z jhover $" )

        parser.add_option("-d", "--debug", 
                          dest="logLevel", 
                          default=logging.WARNING,
                          action="store_const", 
                          const=logging.DEBUG, 
                          help="Set logging level to DEBUG [default WARNING]")
        parser.add_option("-v", "--info", 
                          dest="logLevel", 
                          default=logging.WARNING,
                          action="store_const", 
                          const=logging.INFO, 
                          help="Set logging level to INFO [default WARNING]")
        parser.add_option("--console", 
                          dest="console", 
                          default=False,
                          action="store_true", 
                          help="Forces debug and info messages to be sent to the console")
        parser.add_option("--quiet", dest="logLevel", 
                          default=logging.WARNING,
                          action="store_const", 
                          const=logging.WARNING, 
                          help="Set logging level to WARNING [default]")

        default_conf = "/etc/vc3/vc3-infoservice.conf"
        if 'VC3_SERVICES_HOME' in os.environ:
            default_conf = os.path.join(os.environ['VC3_SERVICES_HOME'], 'etc', 'vc3-infoservice.conf') + ',' + default_conf

        parser.add_option("--conf", dest="confFiles", 
                          default=default_conf,
                          action="store", 
                          metavar="FILE1[,FILE2,FILE3]", 
                          help="Load configuration from FILEs (comma separated list)")

        parser.add_option("--log", dest="logfile", 
                          default="stdout", 
                          metavar="LOGFILE", 
                          action="store", 
                          help="Send logging output to LOGFILE or SYSLOG or stdout [default <syslog>]")
        parser.add_option("--runas", dest="runAs", 
                          #
                          # By default
                          #
                          default=pwd.getpwuid(os.getuid())[0],
                          action="store", 
                          metavar="USERNAME", 
                          help="If run as root, drop privileges to USER")
        (self.options, self.args) = parser.parse_args()

        self.options.confFiles = self.options.confFiles.split(',')

    def __setuplogging(self):
        """ 
        Setup logging 
        """
        self.log = logging.getLogger()
        if self.options.logfile == "stdout":
            logStream = logging.StreamHandler()
        else:
            lf = os.path.expanduser(self.options.logfile)
            logdir = os.path.dirname(lf)
            if not os.path.exists(logdir):
                os.makedirs(logdir)
            runuid = pwd.getpwnam(self.options.runAs).pw_uid
            rungid = pwd.getpwnam(self.options.runAs).pw_gid                  
            os.chown(logdir, runuid, rungid)
            logStream = logging.FileHandler(filename=lf)    

        # Check python version 
        major, minor, release, st, num = sys.version_info
        if major == 2 and minor == 4:
            FORMAT='%(asctime)s (UTC) [ %(levelname)s ] %(name)s %(filename)s:%(lineno)d : %(message)s'
        else:
            FORMAT='%(asctime)s (UTC) [ %(levelname)s ] %(name)s %(filename)s:%(lineno)d %(funcName)s(): %(message)s'
        formatter = logging.Formatter(FORMAT)
        formatter.converter = time.gmtime  # to convert timestamps to UTC
        logStream.setFormatter(formatter)
        self.log.addHandler(logStream)

        # adding a new Handler for the console, 
        # to be used only for DEBUG and INFO modes. 
        if self.options.logLevel in [logging.DEBUG, logging.INFO]:
            if self.options.console:
                console = logging.StreamHandler(sys.stdout)
                console.setFormatter(formatter)
                console.setLevel(self.options.logLevel)
                self.log.addHandler(console)
        self.log.setLevel(self.options.logLevel)
        self.log.info('Logging initialized at level %s.' % self.options.logLevel)


    def _printenv(self):

        envmsg = ''        
        for k in sorted(os.environ.keys()):
            envmsg += '\n%s=%s' %(k, os.environ[k])
        self.log.debug('Environment : %s' %envmsg)


    def __platforminfo(self):
        '''
        display basic info about the platform, for debugging purposes 
        '''
        self.log.info('platform: uname = %s %s %s %s %s %s' %platform.uname())
        self.log.info('platform: platform = %s' %platform.platform())
        self.log.info('platform: python version = %s' %platform.python_version())
        self._printenv()

    def __checkroot(self): 
        """
        If running as root, drop privileges to --runas' account.
        """
        starting_uid = os.getuid()
        starting_gid = os.getgid()
        starting_uid_name = pwd.getpwuid(starting_uid)[0]

        hostname = socket.gethostname()
        
        if os.getuid() != 0:
            self.log.info("Already running as unprivileged user %s at %s" % (starting_uid_name, hostname))
            
        if os.getuid() == 0:
            try:
                runuid = pwd.getpwnam(self.options.runAs).pw_uid
                rungid = pwd.getpwnam(self.options.runAs).pw_gid
                os.chown(self.options.logfile, runuid, rungid)
                
                os.setgid(rungid)
                os.setuid(runuid)
                os.seteuid(runuid)
                os.setegid(rungid)

                self._changehome()
                self._changewd()

                self.log.info("Now running as user %d:%d at %s..." % (runuid, rungid, hostname))
                self._printenv()

            
            except KeyError, e:
                self.log.error('No such user %s, unable run properly. Error: %s' % (self.options.runAs, e))
                sys.exit(1)
                
            except OSError, e:
                self.log.error('Could not set user or group id to %s:%s. Error: %s' % (runuid, rungid, e))
                sys.exit(1)

    def _changehome(self):
        '''
        Set environment HOME to user HOME.
        '''
        runAs_home = pwd.getpwnam(self.options.runAs).pw_dir 
        os.environ['HOME'] = runAs_home
        self.log.debug('Setting up environment variable HOME to %s' %runAs_home)


    def _changewd(self):
        '''
        changing working directory to the HOME directory of the new user,
        '''
        runAs_home = pwd.getpwnam(self.options.runAs).pw_dir
        os.chdir(runAs_home)
        self.log.debug('Switching working directory to %s' %runAs_home)


    def __createconfig(self):
        """Create config, add in options...
        """
        if self.options.confFiles != None:
            try:
                self.config = ConfigParser()
                self.config.read(self.options.confFiles)
            except Exception, e:
                self.log.error('Config failure')
                sys.exit(1)
        
        #self.config.set("global", "configfiles", self.options.confFiles)
           
    def run(self):
        """
        Create Daemon and enter main loop
        """

        try:
            self.log.info('Creating Daemon and entering main loop...')
            infosrv = InfoService(self.config)
            infosrv.run()
            
        except KeyboardInterrupt:
            self.log.info('Caught keyboard interrupt - exitting')
            f.join()
            sys.exit(0)
        except ImportError, errorMsg:
            self.log.error('Failed to import necessary python module: %s' % errorMsg)
            sys.exit(1)
        except:
            self.log.error('''Unexpected exception!''')
            # The following line prints the exception to the logging module
            self.log.error(traceback.format_exc(None))
            print(traceback.format_exc(None))
            sys.exit(1)          

if __name__ == '__main__':
    iscli = InfoServiceCLI()
    iscli.run()




import errno
import json
import logging
import os
import time

class DiskDump(object):
    
    def __init__(self, parent, config, section ):
        self.log = logging.getLogger()
        self.parent = parent
        self.config = config
        self.section = section
        self.documents = {}

        try:
            self.dbname = os.path.expanduser(self.config.get('plugin-diskdump', 'filename', '~/var/infoservice.diskdump'))
        except Exception, e:
            raise e

        self.load_db()
        self.log.debug("DiskDump persistence plugin initialized...")
        
    def storedocument(self, key, doc):
        self.log.debug("Storing doc for key %s..." % key)
        self.documents[key] = doc

        self.store_db()
        return self.documents[key]

    def getdocument(self, key):
        self.load_db()

        self.log.debug("Getting doc for key %s..." % key)
        try:
            s = self.documents[key]
        except KeyError, e:
            s = {}
        return s

    def deletesubtree(self, path):
        self.log.debug("Deleting path %s..." % str(path))

        if len(path) < 1:
            return None

        last_dict = self.documents

        for key in path[0 : -1]:
            last_dict = last_dict[key]

        value = last_dict[path[-1]]
        del last_dict[path[-1]]

        self.store_db()
        return value

    def store_db(self):
        with open(self.dbname, 'w') as outfile:
            outfile.write(json.dumps(self.documents))

    def load_db(self):
        try:
            with open(self.dbname, 'r') as infile:
                self.documents = json.load(infile)
        except IOError, e:
            if e.errno == errno.ENOENT:
                self.log.warn("Could not load db file %s" % self.dbname)
            else:
                raise e
        except ValueError:
            self.log.error("Could not load db file %s" % self.dbname)
            os.rename(self.dbname, self.dbname + '.invalid.' + time.strftime('%Y%m%d.%H%M%S'))
            os.remove(self.dbname)




#! /usr/bin/env python
__author__ = "John Hover, Jose Caballero"
__copyright__ = "2017 John Hover"
__credits__ = []
__license__ = "GPL"
__version__ = "0.9.1"
__maintainer__ = "John Hover"
__email__ = "jhover@bnl.gov"
__status__ = "Production"

import base64
import json
import logging
import logging.handlers
import requests
import urllib
import os
import platform
import sys
import time
import traceback

from random import choice
from string import ascii_uppercase
from optparse import OptionParser
from ConfigParser import ConfigParser, NoOptionError


import requests
requests.packages.urllib3.disable_warnings()
try:
    logging.captureWarnings(True)
except AttributeError:
    # Some versions don't have this. 
    pass

from vc3infoservice.core import InfoEntity

TESTKEY='testkey'
TESTDOC='''{ 
                {"jhoverproject": 
                    { "blueprints": null, 
                      "name": "jhoverproject", 
                      "acl": null, 
                      "members": ["jhover", "angus"], 
                      "state": "new", 
                      "allocations": null, 
                      "owner": "jhover"}
                }
           }'''

class InfoConnectionFailure(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)   

class InfoMissingPairingException(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)  


class InfoClient(object):
    
    def __init__(self, config):
        self.log = logging.getLogger()
        self.log.debug('InfoClient class init...')
        self.config = config

        self.certfile  = False
        self.keyfile   = False
        self.chainfile = False
        try:
            self.certfile  = os.path.expanduser(config.get('netcomm','certfile'))
            self.keyfile   = os.path.expanduser(config.get('netcomm', 'keyfile'))
            self.chainfile = os.path.expanduser(config.get('netcomm','chainfile'))
        except NoOptionError:
            self.log.warning("No complete certificate information was found. Contacting infoservice with unverified certificates.")

        self.httpport  = int(config.get('netcomm','httpport'))
        self.httpsport = int(config.get('netcomm','httpsport'))
        self.infohost  = config.get('netcomm','infohost')
      
        self.log.debug("Client initialized.")

    def storedocument(self, key, doc):
        '''
        Store JSON string <doc> in infoservice under key <key>.
        
        '''
        
        u = "https://%s:%s/info?key=%s" % (self.infohost, 
                            self.httpsport,
                            key
                            )
        self.log.debug("Trying to store document %s at %s" % (doc, u))
        try:
            r = requests.put(u, verify=self.chainfile, cert=(self.certfile, self.keyfile), params={'data' : doc})
            self.log.debug(r.status_code)
        
        except requests.exceptions.ConnectionError, ce:
            self.log.error('Connection failure. %s' % ce)

    def getdocument(self, key):
        '''
        Get and return JSON string for document with key <key>
                
        '''
        u = "https://%s:%s/info?key=%s" % (self.infohost, 
                            self.httpsport,
                            key
                            )
        try:
            r = requests.get(u, verify=self.chainfile, cert=(self.certfile, self.keyfile))
            out = self.stripquotes(r.text)
            parsed = json.loads(out)
            pretty = json.dumps(parsed, indent=4, sort_keys=True)
            #self.log.debug(pretty)
            #self.log.debug(r.status_code)
            return r.text
        
        except requests.exceptions.ConnectionError, ce:
            self.log.error('Connection failure. %s' % ce)
    
    def getdocumentobject(self, key):
        '''
        Get JSON doc and convert to Python and return. 
        
        '''
        u = "https://%s:%s/info?key=%s" % (self.infohost, 
                            self.httpsport,
                            key
                            )
        try:
            r = requests.get(u, verify=self.chainfile, cert=(self.certfile, self.keyfile))
            out = self.stripquotes(r.text)
            parsed = json.loads(out)
            pretty = json.dumps(parsed, indent=4, sort_keys=True)
            return parsed
        
        except requests.exceptions.ConnectionError, ce:
            self.log.error('Connection failure. %s' % ce)       
    
    def storedocumentobject(self, dict, key):
        '''
        Directly store Python dictionary as JSON ...
        
        '''
        if key not in dict.keys():
            td = {}
            td[key] = dict
            dict = td
            
        jstr = json.dumps(dict)
        self.log.debug("JSON string: %s" % jstr)
        self.storedocument(key, jstr)
        
    
    def mergedocument(self, key, doc):
                
        u = "https://%s:%s/info?key=%s" % (self.infohost, 
                            self.httpsport,
                            key
                            )
        self.log.debug("Trying to merge document %s at %s" % (doc, u))
        try:
            r = requests.put(u, verify=self.chainfile, cert=(self.certfile, self.keyfile), params={'data' : doc})
            self.log.debug(r.status_code)
        
        except requests.exceptions.ConnectionError, ce:
            self.log.error('Connection failure. %s' % ce)

    def getbranch(self, *keys):
        doc = self.getdocument(key = keys[0])
        if not doc:
            return None
        ds = json.loads(doc)
        good_keys = []
        for k in keys:
            if k in ds:
                ds = ds[k]
            else:
                raise Exception('Successful keys: ' + str(good_keys))
        return ds
    
    def getsubtree(self, path):
        pass
    
    def deletesubtree(self, path):
        '''
        Delete the leaf given by path.
        
        '''

        try:
            keys = path.split('.')
            key  = keys[0]
        
            u = "https://%s:%s/info?key=%s" % (self.infohost, 
                                self.httpsport,
                                key
                                )
            self.log.debug("Trying to delete document at %s" % (path,))

            r = requests.delete(u, verify=self.chainfile, cert=(self.certfile, self.keyfile), params={'name' : path})
            self.log.debug(r.status_code)
        
        except IndexError, e:
            self.log.error('Path should contain at least one key')
            raise e
        except requests.exceptions.ConnectionError, ce:
            self.log.error('Connection failure. %s' % ce)
        
    def testquery(self):
        self.log.info("Testing storedocument. Doc = %s" % TESTDOC)
        self.storedocument(key=TESTKEY,doc=TESTDOC)
        
        self.log.info("Testing getdocument...")
        self.getdocument(key=TESTKEY)
        self.log.info("Done.")


    def stripquotes(self,s):
        rs = s.replace("'","")
        return rs

    def encode(self, string):
        return base64.b64encode(string)
    
    def decode(self, string):
        return base64.b64decode(string)

    def requestPairing(self, cnsubject):
        '''
        Establish a pairing entry
        '''
        self.log.debug("Infoclient requestPairing for %s " % cnsubject)
        pairingcode = self.generateCode(cnsubject)
        self.log.debug("Generated code: %s " % pairingcode)
        po = Pairing(name=cnsubject, 
                     state='new', 
                     acl=None, 
                     cn=cnsubject, 
                     pairingcode=pairingcode
                     )
        self.log.debug("Made pairing request: %s" % po)
        po.store(self)
        self.log.debug("Stored in /info/pairing..")
        return pairingcode

    def getPairing(self, pairingcode):
        '''
            Special call because it sends pairing code as data, along with URL
        
            :param str pairingcode    Code to be paired with. 
            :return
            :rtype (str, str)         Cert and key
        '''
        u = "https://%s:%s/info?key=pairing&pairingcode=%s" % (self.infohost, 
                            self.httpsport,
                            pairingcode
                            )
        self.log.debug("Attempting to get pairing via URL %s" % u)
        try:
            r = requests.get(u, verify=self.chainfile)     
            pe = json.loads(r.text)
            ecert = pe['cert']
            ekey = pe['key']
            cert = self.decode(ecert)
            key = self.decode(ekey)
            return (cert, key)
       
        except requests.exceptions.ConnectionError, ce:
            self.log.error('Connection failure. %s' % ce)
            raise InfoConnectionFailure("Connection Error.")
        
        except Exception, e:
            self.log.debug('Other failure. Probably missing pairing. %s ' % e)
            raise InfoMissingPairingException("Missing pairing.")
            
            
    def generateCode(self, name ):       
        cs= ''.join(choice(ascii_uppercase) for i in range(6))
        return("%s-%s" % (name, cs))
      

class Pairing(InfoEntity):
    '''
    Represents a request and completed entry for a pairing.
    Stored in /pairing/<code>
    
    '''    
    infokey = 'pairing'
    infoattributes = ['name', 
                      'state',
                      'acl',
                      'cn',   # common name for pair.  
                      'pairingcode',
                      'cert',
                      'key',
                      'encoding'
                     ]
    def __init__(self, name, state, acl, cn, pairingcode, cert=None, key=None, encoding='base64'):
        self.name = name
        self.log = logging.getLogger()
        self.state = state
        self.acl = acl
        self.cn = cn
        self.pairingcode = pairingcode
        self.cert = cert
        self.key = key
        self.encoding = encoding
        

class InfoClientCLI(object):
    """class to handle the command line invocation of APF. 
       parse the input options,
       setup everything, and run Factory class
    """
    def __init__(self):
        self.options = None 
        self.args = None
        self.log = None
        self.config = None

        self.__presetups()
        self.__parseopts()
        self.__setuplogging()
        self.__platforminfo()
        self.__createconfig()
        
    def __presetups(self):
        '''
        we put here some preliminary steps that 
        for one reason or another 
        must be done before anything else
        '''
    
    def __parseopts(self):
        parser = OptionParser(usage='''%prog [OPTIONS]
vc3-info-client is a client for the information store for VC3

This program is licenced under the GPL, as set out in LICENSE file.

Author(s):
John Hover <jhover@bnl.gov>
''', version="%prog $Id: infoclient.py 1-31-17 23:58:06Z jhover $" )

        parser.add_option("-d", "--debug", 
                          dest="logLevel", 
                          default=logging.WARNING,
                          action="store_const", 
                          const=logging.DEBUG, 
                          help="Set logging level to DEBUG [default WARNING]")
        parser.add_option("-v", "--info", 
                          dest="logLevel", 
                          default=logging.WARNING,
                          action="store_const", 
                          const=logging.INFO, 
                          help="Set logging level to INFO [default WARNING]")
        parser.add_option("--quiet", dest="logLevel", 
                          default=logging.WARNING,
                          action="store_const", 
                          const=logging.WARNING, 
                          help="Set logging level to WARNING [default]")

        default_conf = "~/etc/vc3/vc3-infoclient.conf"
        if 'VC3_SERVICES_HOME' in os.environ:
            default_conf = os.path.join(os.environ['VC3_SERVICES_HOME'], 'etc', 'vc3-infoclient.conf') + ',' + default_conf

        parser.add_option("-c","--config", dest="confFiles", 
                          default=default_conf,
                          action="store", 
                          metavar="FILE1[,FILE2,FILE3]", 
                          help="Load configuration from FILEs (comma separated list)")

        parser.add_option("--log", dest="logfile", 
                          default="stdout", 
                          metavar="LOGFILE", 
                          action="store", 
                          help="Send logging output to LOGFILE or SYSLOG or stdout [default <syslog>]")
        parser.add_option("--add", dest="addfiles", 
                          action="store", 
                          metavar="FILE1[,FILE2,FILE3]", 
                          help="Merge info into store from JSON files with key as top-level tag.")        
        parser.add_option("--getkey", dest="getkey", 
                          action="store", 
                          metavar="[resource|account|cluster|...]", 
                          help="Get info from store with provided key.")        
        parser.add_option("--deletesubtree", dest="deletesubtree", 
                          action="store", 
                          metavar="[resource|account|cluster|...]", 
                          help="Delete the leaf given by keys path. Path given in the form of a.b.c.d... ")        

        parser.add_option("--requestpairing",
                          dest = "requestpairing",
                          action="store",
                          metavar="SUBJECT/CN",
                          help="Request a cert/key creation for supplied HOST/CN/SUBJECT"
                          )
        
        parser.add_option("--getpairing",
                          dest = "pairingcode",
                          action = "store",
                          metavar="PAIRINGCODE",
                          help="Retrieve cert/key pair for supplied pairing code."
                          )
                        
        (self.options, self.args) = parser.parse_args()
        self.options.confFiles = self.options.confFiles.split(',')

    def __setuplogging(self):
        """ 
        Setup logging 
        """
        self.log = logging.getLogger()
        if self.options.logfile == "stdout":
            logStream = logging.StreamHandler()
        else:
            lf = self.options.logfile
            logdir = os.path.dirname(lf)
            if not os.path.exists(logdir):
                os.makedirs(logdir)
            logStream = logging.FileHandler(filename=lf)    

        # Check python version 
        major, minor, release, st, num = sys.version_info
        if major == 2 and minor == 4:
            FORMAT='%(asctime)s (UTC) [ %(levelname)s ] %(name)s %(filename)s:%(lineno)d : %(message)s'
        else:
            FORMAT='%(asctime)s (UTC) [ %(levelname)s ] %(name)s %(filename)s:%(lineno)d %(funcName)s(): %(message)s'
        formatter = logging.Formatter(FORMAT)
        formatter.converter = time.gmtime  # to convert timestamps to UTC
        logStream.setFormatter(formatter)
        self.log.addHandler(logStream)

        self.log.setLevel(self.options.logLevel)
        self.log.info('Logging initialized.')


    def _printenv(self):

        envmsg = ''        
        for k in sorted(os.environ.keys()):
            envmsg += '\n%s=%s' %(k, os.environ[k])
        self.log.debug('Environment : %s' %envmsg)


    def __platforminfo(self):
        '''
        display basic info about the platform, for debugging purposes 
        '''
        self.log.info('platform: uname = %s %s %s %s %s %s' %platform.uname())
        self.log.info('platform: platform = %s' %platform.platform())
        self.log.info('platform: python version = %s' %platform.python_version())
        self._printenv()

    
    def __createconfig(self):
        """Create config, add in options...
        """
        if self.options.confFiles != None:
            try:
                self.config = ConfigParser()
                self.config.read(self.options.confFiles)
            except Exception, e:
                self.log.error('Config failure')
                sys.exit(1)
        
        #self.config.add_section('cli')
        #self.config.set("cli", "addfiles", self.options.confFiles)

    def doquery(self):
        self.ic = InfoClient(self.config)
        if self.options.addfiles:
            flist = self.options.addfiles.split(',')
            for fn in flist:
                fname = fn.strip()
                self.log.debug("Adding contents of file %s" % fname)
                jdoc = open(fname).read()
                data = json.loads(jdoc)
                pretty = json.dumps(data, indent=4, sort_keys=True)
                self.log.debug(pretty)
                k = data.keys()[0]
                self.log.debug("key is %s" % k)
                self.ic.mergedocument(k,jdoc)
        
        if self.options.getkey:
            qkey = self.options.getkey.lower().strip()
            self.log.debug("Getkey is %s, doing query" % qkey )
            out = self.ic.getdocument(qkey)
            print(out)

        if self.options.deletesubtree:
            dpath = self.options.deletesubtree.lower().strip()
            self.log.debug("Deletesubtree is %s, doing query" % dpath )
            out = self.ic.deletesubtree(dpath)
            print(out)
            
        if self.options.requestpairing:
            self.log.debug("Setting up pairing for CN: %s"% self.options.requestpairing )
            code = self.ic.requestPairing(self.options.requestpairing)
            print("%s" % code)
        
        if self.options.pairingcode:
            try:
                #out = self.ic.getPairing(self.options.pairingcode)
                #print("out is %s" % out)
                (cert, key) = self.ic.getPairing(self.options.pairingcode)
                print("%s" % cert)
                print("")
                print("%s" % key)
            except Exception, e:
                self.log.error("Exception %s" % e)
                print("No response. Invalid pairing or not setup yet. Try in 30 seconds.")

    def testquery(self):
        self.ic = InfoClient(self.config)
        self.ic.testquery()
        

    def run(self):
        self.doquery()


if __name__ == '__main__':
    logging.info("Running from .py file...")
    iccli = InfoClientCLI()
    iccli.run()




#!/usr/bin/env python
#
# Setup prog for infoservice

import sys
import re
import time
from setuptools import setup


def choose_data_file_locations():
    local_install = False

    if '--user' in sys.argv:
        local_install = True
    elif any([re.match('--home(=|\s)', arg) for arg in sys.argv]):
        local_install = True
    elif any([re.match('--prefix(=|\s)', arg) for arg in sys.argv]):
        local_install = True

    if local_install:
        return home_data_files
    else:
        return rpm_data_files

current_time = time.gmtime()
release_version = "{0}.{1:0>2}.{2:0>2}".format(current_time.tm_year, current_time.tm_mon, current_time.tm_mday)

systemd_files = ['etc/vc3-infoservice.service']

etc_files = ['etc/vc3-infoservice.conf',
             'etc/vc3-infoclient.conf',]

sysconfig_files = ['etc/sysconfig/vc3-infoservice', ]

logrotate_files = ['etc/logrotate/vc3-infoservice', ]

initd_files = ['etc/vc3-infoservice.init', ]

rpm_data_files = [('/etc/vc3', etc_files),
                  ('/etc/sysconfig', sysconfig_files),
                  ('/etc/logrotate.d', logrotate_files),
                  ('/etc/init.d', initd_files),
                  ('/usr/lib/systemd/system', systemd_files), ]

home_data_files = [('etc', etc_files),
                   ('etc', initd_files),
                   ('etc', sysconfig_files), ]

data_files = choose_data_file_locations()

# ===========================================================

setup(
    name="vc3-info-service",
    version=release_version,
    description='vc3-info-service package',
    long_description='''This package contains the VC3 Information Service''',
    license='GPL',
    author='John Hover',
    author_email='jhover@bnl.gov',
    maintainer='John Hover',
    maintainer_email='jhover@bnl.gov',
    url='https://github.com/vc3-project',
    packages=['vc3infoservice',
              'vc3infoservice.plugins',
              'vc3infoservice.plugins.persist'
              ],
    scripts=['scripts/vc3-info-service',
             'scripts/vc3-info-client',
             ],
    data_files=data_files,
    install_requires=['cherrypy','pyopenssl','pluginmanager']
)




'''

For CouchDB our config entity categories map nicely onto the JSON document databases. 
Use them directly...

'''


import logging

class CouchDB(object):
    
    def __init__(self, parent, config, section ):
        self.log = logging.getLogger()
        self.parent = parent
        self.config = config
        self.section = section
        self.log.debug("CouchDB persistence plugin initialized...")
        
    def storedocument(self, key, doc):
        self.log.debug("Storing doc for key %s..." % key)
        
        
        
    def getdocument(self, key):
        self.log.debug("Getting doc for key %s..." % key)
        

        


import logging

class Memory(object):
    
    def __init__(self, parent, config, section ):
        self.log = logging.getLogger()
        self.parent = parent
        self.config = config
        self.section = section
        self.documents = {}
        self.log.debug("Memory persistence plugin initialized...")
        
    def storedocument(self, key, doc):
        self.log.debug("Storing doc for key %s..." % key)
        self.documents[key] = doc

    def deletesubtree(self, path):
        self.log.debug("Deleting path %s..." % str(path))

        if len(path) < 1:
            return None

        last_dict = self.documents

        for key in path[0 : -1]:
            last_dict = last_dict[key]

        value = last_dict[path[-1]]
        del last_dict[path[-1]]

        return value
        
    def getdocument(self, key):
        self.log.debug("Getting doc for key %s..." % key)
        try:
            s = self.documents[key]
        except KeyError, e:
            s = {}
        return s
        
        
        
